
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




import os, os.path
import ntpath
import logging
import random
import librosa

from accessify                  import private, protected
from modules.base.enum          import PlaylistType
from modules.base.utils         import SimpleUtil, EngineUtil
from modules.communication.mail import AuraMailer
from modules.base.enum          import ChannelType


class FallbackManager:
    """
    Handles all types of fallbacks in case there is an outage
    for the regular radio programme.

    Attributes:
        config (AuraConfig):        The engine configuration
        logger (AuraLogger):        The logger
        mail   (AuraMailer):        Mail service
        scheduler (AuraScheduler):  The scheduler
        fallback_history (Dict):    Holds a 24h history of played, local tracks to avoid re-play
        last_fallback (Integer):    Timestamp, when the last local file fallback was played
        is_processing (Boolean):    Flag to avoid race-conditions, as Liquidsoap sends plenty of requests at once
    """

    config = None
    logger = None
    mailer = None
    scheduler = None
    fallback_history = {}
    last_fallback = 0
    is_processing = False


    def __init__(self, config, logger, scheduler):
        """
        Constructor

        Args:
            config (AuraConfig):    Holds the engine configuration
        """
        self.config = config
        self.mailer = AuraMailer(self.config)
        self.scheduler = scheduler
        self.logger = logger


    #
    #   PUBLIC METHODS
    #

    def resolve_playlist(self, schedule):
        """
        Resolves the (fallback) playlist for the given schedule in case of pro-active fallback scenarios.
        
        A resolved playlist represents the state how it would currently  be aired. For example the `FallbackManager` 
        evaluated, that the actually planned playlist cannot be played for various reasons (e.g. entries n/a). 
        Instead one of the fallback playlists should be played. If the method is called some time later,
        it actually planned playlist might be valid, thus returned as the resolved playlist.

        As long the adressed schedule is still within the scheduling window, the resolved playlist can
        always change.

        This method also updates `schedule.fallback_state` to the current fallback type (`PlaylistType`).

        Args:
            schedule (Schedule):    The schedule to resolve the playlist for

        
        Returns:
            (Playlist):             The resolved playlist
        """
        playlist = None
        type = None
        self.logger.info("Resolving playlist for schedule #%s ..." % schedule.schedule_id)

        if not self.validate_playlist(schedule, "playlist"):
            if not self.validate_playlist(schedule, "schedule_fallback"):
                if not self.validate_playlist(schedule, "show_fallback"):
                    if not self.validate_playlist(schedule, "station_fallback"):
                        raise Exception("No (fallback) playlists for schedule #%s available - not even a single one!" % schedule.schedule_id)
                    else:
                        type = PlaylistType.STATION
                        playlist = schedule.station_fallback
                else:
                    type = PlaylistType.TIMESLOT
                    playlist = schedule.schedule_fallback
            else:
                type = PlaylistType.SHOW
                playlist = schedule.show_fallback
        else:
            type = PlaylistType.DEFAULT
            playlist = schedule.playlist

        if type and type != PlaylistType.DEFAULT:
            previous_type = schedule.fallback_state
            if type == previous_type:
                self.logger.info("Fallback state for schedule #%s is still '%s'" % (schedule.schedule_id, type))
            else:
                self.logger.warn("Detected fallback type switch from '%s' to '%s' is required for schedule %s." % (previous_type, type, str(schedule)))
        
        schedule.fallback_state = type
        return playlist[0]



    def handle_proactive_fallback(self, scheduler, playlist):
        """
        This is the 1st level strategy for fallback handling. When playlist entries are pre-rolled their 
        state is validated. If any of them doesn't become "ready to play" in time, some fallback entries
        are queued.
        """
        resolved_playlist = self.resolve_playlist(playlist.schedule)
        if playlist != resolved_playlist:
            self.logger.info("Switching from playlist #%s to fallback playlist #%s ..." % (playlist.playlist_id, resolved_playlist.playlist_id))
            
            # Destroy any existing queue timers
            for entry in playlist.entries:
                scheduler.stop_timer(entry.switchtimer)
            self.logger.info("Stopped existing timers for entries")

            # Queue the fallback playlist
            scheduler.queue_playlist_entries(resolved_playlist.schedule, resolved_playlist.entries, False, True)
            self.logger.info("Queued fallback playlist entries (Fallback type: %s)" % playlist.type)
        else:
            self.logger.critical(SimpleUtil.red("For some strange reason the fallback playlist equals the currently failed one?!"))

        


    def get_fallback_for(self, fallbackname):
        """
        Retrieves a random fallback audio source for any of the types:
            - timeslot/schedule
            - show
            - station
        
        Args:
            fallbackname (String):      Fallback type

        Returns:
            (String):                   Absolute path to the file
        """
        file = ""
        media_type = "PLAYLIST"
        active_schedule, active_playlist = self.scheduler.get_active_playlist()

        # Block access to avoid race-conditions
        if self.is_processing:
            return None
        else:
            self.is_processing = True

        # Get fallback track(s) by fallback-type
        if fallbackname == "timeslot":
            file = self.get_playlist_items(active_schedule, "schedule_fallback")

        elif fallbackname == "show":
            file = self.get_playlist_items(active_schedule, "show_fallback")

        elif fallbackname == "station":
            file = self.get_playlist_items(active_schedule, "station_fallback")

            if not file:
                media_type = "TRACK"
                file = self.get_random_local_track()

                if not file:
                    self.logger.critical("Got no file for station fallback! Playing default test track, to play anything at all.")
                    file = "../../testing/content/ernie_mayne_sugar.mp3"
                    media_type = "DEFAULT TRACK"
        else:
            file = ""
            self.logger.critical("Should set next fallback file for " + fallbackname + ", but this fallback is unknown!")

        if file:
            # Send admin email to notify about the fallback state
            if not active_playlist:
                active_playlist = "-"
            msg = "AURA ENGINE %s FALLBACK DETECTED!\n\n" % fallbackname
            msg += "Expected, active Schedule: %s \n" % active_schedule
            msg += "Expected, active Playlist: %s \n\n" % active_playlist
            msg += "Providing FALLBACK-%s for %s '%s'\n\n" % (media_type, fallbackname, file)
            msg += "Please review the schedules or contact your Aura Engine administrator."
            self.mailer.send_admin_mail("CRITICAL - Detected fallback for %s" % fallbackname, msg)
            self.logger.warn("Providing fallback %s: '%s'. Sent admin email about fallback state" % (media_type, file))

        self.is_processing = False
        return file



    def fallback_has_started(self, artist, title):
        """
        Called when a fallback track has actually started playing
        """
        self.logger.info("Now playing: fallback track '%s - %s'." % (artist, title))



    def get_track_duration(self, file):
        """
        Returns the length of the given audio file.
        """
        y, sr = librosa.load(file)
        duration = librosa.get_duration(y=y, sr=sr)
        return duration



    #
    #   PRIVATE METHODS
    #



    def validate_playlist(self, schedule, playlist_type):
        """
        Checks if a playlist is valid for play-out.
        """
        playlist = getattr(schedule, playlist_type)
        if playlist \
            and isinstance(playlist, list) \
            and playlist[0].entries \
            and len(playlist[0].entries) > 0:

            return self.validate_entries(playlist[0].entries)
        return False



    def validate_entries(self, entries):
        """
        Checks if playlist entries are valid for play-out.
        """
        for entry in entries:
            if entry.get_type() == ChannelType.FILESYSTEM:
                audio_store = self.config.get("audiofolder")
                filepath = EngineUtil.uri_to_filepath(audio_store, entry.source)

                if not self.is_audio_file(filepath):
                    self.logger.warn("Invalid filesystem path '%s' in entry '%s'" % (filepath, str(entry)))
                    return False
        return True



    def get_playlist_items(self, schedule, fallback_key):
        """
        Retrieves the list of tracks from a playlist defined by `fallback_key`.
        """
        playlist_files = ""

        if hasattr(schedule, fallback_key):
            playlist = getattr(schedule, fallback_key)
            if len(playlist) > 0:
                playlist = playlist[0]
                if playlist and playlist.entries:
                    for entry in playlist.entries:
                        playlist_files += entry.source + "\n"

        return playlist_files



    def get_random_local_track(self):
        """
        Retrieves a random audio track from the local station-fallback directory.

        Returns:
            (String):   Absolute path to an audio file
        """
        dir = self.config.fallback_music_folder
        files = os.listdir(dir)
        audio_files = list(filter(lambda f: self.is_audio_file(os.path.join(dir, f)), files))
        
        if not dir or not audio_files:
            self.logger.error("Folder 'fallback_music_folder = %s' is empty!" % dir)
            return None

        # If last played fallback is > 24 hours ago, ignore play history
        # This should save used memory if the engine runs for a long time
        if self.last_fallback < SimpleUtil.timestamp() - (60*60*24):
            self.fallback_history = {}
            self.logger.info("Cleared fallback history.")
        self.last_fallback = SimpleUtil.timestamp()
        
        # Retrieve files which haven't been played yet
        history = set(self.fallback_history.keys())
        left_audio_files = list( set(audio_files) - (history) )
        self.logger.info("Left fallback audio-files: %d/%d" % (len(left_audio_files), len(audio_files)))
        
        # If nothing left, clear history and start with all files again
        if not len(left_audio_files):
            self.fallback_history = {}
            left_audio_files = audio_files

        # Select random track from directory
        i = random.randint(0, len(left_audio_files)-1)
        file = os.path.join(dir, left_audio_files[i])

        # Store track in history, to avoid playing it multiple times
        if file:
            self.fallback_history[left_audio_files[i]] = SimpleUtil.timestamp()

        return file



    def is_audio_file(self, file):
        """
        Checks if the passed file is an audio file i.e. has a file-extension
        known for audio files.

        Args:
            dir (String):  
            file (File):   the file object.

        Returns:
            (Boolean):      True, if it's an audio file.
        """
        audio_extensions = [".wav", ".flac", ".mp3", ".ogg", ".m4a"]
        ext = os.path.splitext(file)[1]

        if os.path.isfile(file):
            if any(ext in s for s in audio_extensions):
                return True
        return False